from kinto.core.events import ResourceChanged, ACTIONS
from kinto.core.errors import http_error
from pyramid.events import subscriber
from pyramid.exceptions import HTTPBadRequest
from pyramid.settings import asbool

from . import utils


# XXX: Use AfterResourceChanged when implemented.
@subscriber(ResourceChanged,
            for_resources=('record', 'collection', 'bucket'),
            for_actions=(ACTIONS.DELETE,))
def on_delete_record(event):
    """When a resource record is deleted, delete all related attachments.
    When a bucket or collection is deleted, it removes the attachments of
    every underlying records.
    """
    keep_old_files = asbool(utils.setting_value(event.request, 'keep_old_files', default=False))

    # Retrieve attachments for these records using links.
    resource_name = event.payload['resource_name']
    filter_field = '%s_uri' % resource_name
    uri = event.payload['uri']
    utils.delete_attachment(event.request, link_field=filter_field, uri=uri,
                            keep_old_files=keep_old_files)


@subscriber(ResourceChanged,
            for_resources=('record',),
            for_actions=(ACTIONS.UPDATE,))
def on_update_record(event):
    if getattr(event.request, '_attachment_auto_save', False):
        # Record attributes are being updated by the plugin itself.
        return

    # A user is changing the record, make sure attachment metadata is not
    # altered manually.
    for change in event.impacted_records:
        attachment_before = change['old'].get('attachment')
        attachment_after = change['new'].get('attachment')
        if attachment_before and attachment_after:
            if attachment_before != attachment_after:
                error_msg = "Attachment metadata cannot be modified."
                raise http_error(HTTPBadRequest(), message=error_msg)



import uuid
import os

from urllib.parse import urlparse, urlencode, parse_qsl, urlunparse

import webtest
from kinto.core import utils as core_utils
from kinto.core import testing as core_support

from pyramid_storage.s3 import S3FileStorage
from pyramid_storage.interfaces import IFileStorage


SAMPLE_SCHEMA = {
    "title": "Font file",
    "type": "object",
    "properties": {
        "family": {"type": "string"},
        "author": {"type": "string"},
    }
}


def build_url(url, **params):
    url_parts = list(urlparse(url))
    query = dict(parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urlencode(query)
    return urlunparse(url_parts)


def get_user_headers(user):
    credentials = "%s:secret" % user
    authorization = 'Basic {0}'.format(core_utils.encode64(credentials))
    return {
        'Authorization': authorization
    }


class BaseWebTest(object):
    config = ''

    def __init__(self, *args, **kwargs):
        super(BaseWebTest, self).__init__(*args, **kwargs)
        self.app = self.make_app()
        self.backend = self.app.app.registry.getUtility(IFileStorage)
        self.base_url = self.backend.url('')
        self._created = []

    def setUp(self):
        super(BaseWebTest, self).setUp()
        self.headers = {
            'Content-Type': 'application/json',
            'Origin': 'http://localhost:9999'
        }
        self.headers.update(get_user_headers('mat'))

        self.create_collection('fennec', 'fonts')
        self.record_id = _id = str(uuid.uuid4())
        self.record_uri = self.get_record_uri('fennec', 'fonts', _id)
        self.endpoint_uri = self.record_uri + '/attachment'
        self.default_files = [('attachment', 'image.jpg', b'--fake--')]
        self.file_field = 'attachment'

    @property
    def nb_uploaded_files(self):
        return len(self.default_files)

    def make_app(self):
        curdir = os.path.dirname(os.path.realpath(__file__))
        app = webtest.TestApp("config:%s" % self.config, relative_to=curdir)
        app.RequestClass = core_support.get_request_class(prefix="v1")
        return app

    def upload(self, params=[], files=None, headers={}, status=None):
        files = files or self.default_files
        headers = headers or self.headers.copy()
        content_type, body = self.app.encode_multipart(params, files)
        headers['Content-Type'] = content_type

        endpoint_url = self.endpoint_uri

        resp = self.app.post(endpoint_url, body, headers=headers,
                             status=status)
        if 200 <= resp.status_code < 300:
            self._add_to_cleanup(resp.json)

        return resp

    def _add_to_cleanup(self, attachment):
        relativeurl = attachment['location'].replace(self.base_url, '')
        self._created.append(relativeurl)

    def create_collection(self, bucket_id, collection_id):
        bucket_uri = '/buckets/%s' % bucket_id
        self.app.put_json(bucket_uri,
                          {},
                          headers=self.headers)
        collection_uri = bucket_uri + '/collections/%s' % collection_id
        collection = {
            'schema': SAMPLE_SCHEMA
        }
        self.app.put_json(collection_uri,
                          {'data': collection},
                          headers=self.headers)

    def get_record_uri(self, bucket_id, collection_id, record_id):
        return ('/buckets/{bucket_id}/collections/{collection_id}'
                '/records/{record_id}').format(**locals())

    def get_record(self, resp):
        # Alias to resp.json, in a separate method to easily be extended.
        return resp.json


class BaseWebTestLocal(BaseWebTest):
    config = 'config/local.ini'

    def tearDown(self):
        """Delete uploaded local files.
        """
        super(BaseWebTest, self).tearDown()
        basepath = self.app.app.registry.settings['kinto.attachment.base_path']
        for created in self._created:
            filepath = os.path.join(basepath, created)
            if os.path.exists(filepath):
                os.remove(filepath)


class BaseWebTestS3(BaseWebTest):
    config = 'config/s3.ini'

    def __init__(self, *args, **kwargs):
        self._s3_bucket_created = False
        super(BaseWebTestS3, self).__init__(*args, **kwargs)

    def make_app(self):
        app = super(BaseWebTestS3, self).make_app()

        # Create the S3 bucket if necessary
        if not self._s3_bucket_created:
            prefix = 'kinto.attachment.'
            settings = app.app.registry.settings
            fs = S3FileStorage.from_settings(settings, prefix=prefix)

            bucket_name = settings[prefix + 'aws.bucket_name']
            fs.get_connection().create_bucket(bucket_name)
            self._s3_bucket_created = True

        return app



import pkg_resources
from collections import defaultdict
from pyramid.exceptions import ConfigurationError
from pyramid.settings import asbool

#: Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution(__package__).version


def includeme(config):
    # Process settings to remove storage wording.
    settings = config.get_settings()

    storage_settings = {}
    config.registry.attachment_resources = defaultdict(dict)

    for setting_name, setting_value in settings.items():
        if setting_name.startswith('attachment.'):
            if setting_name.startswith('attachment.resources.'):
                # Resource specific config
                parts = setting_name.replace('attachment.resources.', '').split('.')
                # attachment.resources.{bid}.gzipped
                # attachment.resources.{bid}.{cid}.gzipped
                if len(parts) == 3:
                    bucket_id, collection_id, name = parts
                    resource_id = '/buckets/{}/collections/{}'.format(bucket_id, collection_id)
                elif len(parts) == 2:
                    bucket_id, name = parts
                    resource_id = '/buckets/{}'.format(bucket_id)
                else:
                    message = 'Configuration rule malformed: `{}`'.format(setting_name)
                    raise ConfigurationError(message)

                if name in ('gzipped', 'randomize', 'keep_old_files'):
                    config.registry.attachment_resources[resource_id][name] = asbool(setting_value)
                else:
                    message = '`{}` is not a supported setting name. Read `{}`'.format(
                        name, setting_name)
                    raise ConfigurationError(message)
            else:
                setting_name = setting_name.replace('attachment.', 'storage.')
                storage_settings[setting_name] = setting_value

    # Force some pyramid_storage settings.
    storage_settings['storage.name'] = 'attachment'
    storage_settings.setdefault('storage.extensions', 'default')
    config.add_settings(storage_settings)

    # It may be useful to define an additional base_url setting.
    # (see workaround about relative base_url in README)
    extra_base_url = settings.get('attachment.extra.base_url',
                                  settings.get('attachment.base_url'))

    # Expose capability.
    config.add_api_capability("attachments",
                              version=__version__,
                              description="Add file attachments to records",
                              url="https://github.com/Kinto/kinto-attachment/",
                              base_url=extra_base_url)

    # Register heartbeat to check attachments storage.
    from kinto_attachment.views import attachments_ping
    config.registry.heartbeats['attachments'] = attachments_ping

    # Enable attachment backend.
    if 'storage.base_path' in storage_settings:
        config.include('pyramid_storage.local')
    else:
        config.include('pyramid_storage.s3')

    config.scan('kinto_attachment.views')
    config.scan('kinto_attachment.listeners')



import gzip
import hashlib
import os

import requests
from kinto_http import cli_utils


def sha256(content):
    m = hashlib.sha256()
    m.update(content)
    return m.hexdigest()


def download_files(client, records, folder, chunk_size=1024):

    for record in records:
        if 'attachment' not in record:
            continue

        attachment = record['attachment']
        # Check if file was Gzipped during upload (see `upload.py`)
        is_gzip = 'original' in attachment
        if is_gzip:
            filename = attachment['original']['filename']
            remote_hash = attachment['original']['hash']
        else:
            filename = attachment['filename']
            remote_hash = attachment['hash']

        destination = os.path.join(folder, filename)

        # Compare local hash with remote and skip if equal.
        if os.path.exists(destination):
            local_hash = sha256(open(destination, 'rb').read())
            if local_hash == remote_hash:
                print('Skip "%s". Up-to-date.' % filename)
                continue

        # Download remote attachment by chunk.
        resp = requests.get(attachment['location'], stream=True)
        resp.raise_for_status()
        tmp_file = destination + '.tmp'
        with open(tmp_file, 'wb') as f:
            for chunk in resp.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)

        # Decompress the file if necessary.
        if is_gzip:
            with open(destination, 'wb') as f:
                data = open(tmp_file, 'rb').read()
                f.write(gzip.decompress(data))
        else:
            os.rename(tmp_file, destination)

        print('Downloaded "%s"' % filename)


def main():
    parser = cli_utils.add_parser_options(
        description='Download files from Kinto')
    parser.add_argument('-f', '--folder', help='Folder to download files in.',
                        type=str, default=".")
    args = parser.parse_args()

    client = cli_utils.create_client_from_args(args)

    # See if timestamp was saved from last run.
    last_sync = None
    timestamp_file = os.path.join(args.folder, '.last_sync')
    if os.path.exists(args.folder):
        if os.path.exists(timestamp_file):
            last_sync = open(timestamp_file, 'r').read()
    else:
        os.makedirs(args.folder)

    # Retrieve the collection of records.
    existing = client.get_records(_since=last_sync, _sort="-last_modified")

    if existing:
        download_files(client, existing, args.folder)

        timestamp = max([r['last_modified'] for r in existing])
        # Save the highest timestamp for next runs.
        with open(timestamp_file, 'w') as f:
            f.write("%s" % timestamp)


if __name__ == '__main__':
    main()



import pprint

from kinto_http import cli_utils

DEFAULT_SERVER = "http://localhost:8888/v1"


def main():
    parser = cli_utils.add_parser_options(
        description='Create account on Kinto',
        default_server=DEFAULT_SERVER)

    args = parser.parse_args()

    user, password = args.auth

    client = cli_utils.create_client_from_args(args)
    root_url = client.get_endpoint('root')
    account_url = root_url + "/accounts/" + user
    resp, _ = client.session.request(method='PUT', endpoint=account_url, payload={
        "data": {"password": password}
    })
    pprint.pprint(resp)


if __name__ == '__main__':
    main()



import cgi
import json
from io import BytesIO

from kinto.core import logger, Service
from kinto.core.authorization import DYNAMIC as DYNAMIC_PERMISSION
from kinto.core.errors import ERRORS, http_error
from pyramid import httpexceptions
from pyramid.settings import asbool

from kinto_attachment import utils


HEARTBEAT_CONTENT = '{"test": "write"}'
HEARTBEAT_FILENAME = 'heartbeat'
SINGLE_FILE_FIELD = 'attachment'


attachment = Service(name='attachment',
                     description='Attach a file to a record',
                     path=utils.RECORD_PATH + '/attachment',
                     factory=utils.AttachmentRouteFactory)


@attachment.post(permission=DYNAMIC_PERMISSION)
def attachment_post(request):
    return post_attachment_view(request, SINGLE_FILE_FIELD)


@attachment.delete(permission=DYNAMIC_PERMISSION)
def attachment_delete(request):
    return delete_attachment_view(request, SINGLE_FILE_FIELD)


def post_attachment_view(request, file_field):
    keep_old_files = asbool(utils.setting_value(request, 'keep_old_files', default=False))

    # Remove potential existing attachment.
    utils.delete_attachment(request, keep_old_files=keep_old_files)

    if "multipart/form-data" not in request.headers.get('Content-Type', ''):
        raise http_error(httpexceptions.HTTPBadRequest(),
                         errno=ERRORS.INVALID_PARAMETERS,
                         message="Content-Type should be multipart/form-data")

    # Store file locally.
    try:
        content = request.POST.get(file_field)
    except ValueError as e:
        raise http_error(httpexceptions.HTTPBadRequest(),
                         errno=ERRORS.INVALID_PARAMETERS.value,
                         message=str(e))

    if content is None:
        raise http_error(httpexceptions.HTTPBadRequest(),
                         errno=ERRORS.INVALID_POSTED_DATA,
                         message="Attachment missing.")

    folder_pattern = utils.setting_value(request, 'folder', default='')
    folder = folder_pattern.format(**request.matchdict) or None
    attachment = utils.save_file(request, content, folder=folder)

    # Update related record.
    posted_data = {k: v for k, v in request.POST.items() if k != file_field}
    record = {'data': {}}
    for field in ('data', 'permissions'):
        if field in posted_data:
            try:
                record[field] = json.loads(posted_data.pop(field))
            except ValueError as e:
                error_msg = "body: %s is not valid JSON (%s)" % (field, str(e))
                raise http_error(httpexceptions.HTTPBadRequest(),
                                 errno=ERRORS.INVALID_POSTED_DATA,
                                 message=error_msg)
    # Some fields remaining in posted_data after popping: invalid!
    for field in posted_data.keys():
        error_msg = "body: %r not in ('data', 'permissions')" % field
        raise http_error(httpexceptions.HTTPBadRequest(),
                         errno=ERRORS.INVALID_POSTED_DATA,
                         message=error_msg)

    record['data'][file_field] = attachment

    utils.patch_record(record, request)

    # Return attachment data (with location header)
    request.response.headers['Location'] = utils.record_uri(request,
                                                            prefix=True)
    return attachment


def delete_attachment_view(request, file_field):
    keep_old_files = asbool(utils.setting_value(request, 'keep_old_files', default=False))

    utils.delete_attachment(request, keep_old_files=keep_old_files)

    # Remove metadata.
    record = {"data": {}}
    record["data"][file_field] = None
    utils.patch_record(record, request)

    request.response.status = 204


def attachments_ping(request):
    """Heartbeat view for the attachments backend.
    :returns: ``True`` if succeeds to write and delete, ``False`` otherwise.
    """
    # Do nothing if server is readonly.
    if asbool(request.registry.settings.get('readonly', False)):
        return True

    # We will fake a file upload, so pick a file extension that is allowed.
    extensions = request.attachment.extensions or {'json'}
    allowed_extension = "." + list(extensions)[-1]

    status = False
    try:
        content = cgi.FieldStorage()
        content.filename = HEARTBEAT_FILENAME + allowed_extension
        content.file = BytesIO(HEARTBEAT_CONTENT.encode('utf-8'))
        content.type = 'application/octet-stream'

        stored = utils.save_file(request, content, keep_link=False, replace=True)

        relative_location = stored['location'].replace(request.attachment.base_url, '')
        request.attachment.delete(relative_location)

        status = True
    except Exception as e:
        logger.exception(e)
    return status



import cgi
import unittest
from io import BytesIO

from kinto_attachment.utils import save_file


class _Registry(object):
    settings = {'attachment.folder': ''}
    attachment_resources = {}

    def save(self, *args, **kw):
        return 'yeahok'

    def url(self, location):
        return 'http://localhost/%s' % location

    def create(self, *args, **kw):
        pass

    @property
    def storage(self):
        return self


class _Request(object):
    registry = _Registry()
    matchdict = {'bucket_id': 'bucket',
                 'collection_id': 'collection'}

    attachment = _Registry()

    def route_path(self, *args, **kw):
        return 'fullpath'


class TestUtils(unittest.TestCase):

    def test_save_file_gzip(self):
        my_font = cgi.FieldStorage()
        my_font.filename = 'font.ttf'
        my_font.file = BytesIO(b'content')
        my_font.type = 'application/x-font'

        request = _Request()
        request.registry.settings['attachment.gzipped'] = True
        res = save_file(request, my_font)
        self.assertTrue('original' in res)

    def test_save_file_not_gzip(self):
        my_font = cgi.FieldStorage()
        my_font.filename = 'font.ttf'
        my_font.file = BytesIO(b'content')
        my_font.type = 'application/x-font'

        request = _Request()
        request.registry.settings['attachment.gzipped'] = False
        res = save_file(request, my_font)
        self.assertFalse('original' in res)



import unittest
from . import BaseWebTest


class Listener(object):
    def __init__(self):
        self.received = []

    def __call__(self, event):
        self.received.append(event)


listener = Listener()


def load_from_config(config, prefix):
    return listener


class ResourceChangedTest(BaseWebTest, unittest.TestCase):
    config = 'config/events.ini'

    def test_resource_changed_is_triggered_when_attachment_is_set(self):
        before = len(listener.received)
        self.upload()
        self.assertEqual(len(listener.received), before + 1)

    def test_action_is_create_or_update(self):
        self.upload()
        self.assertEqual(listener.received[-1].payload['action'], 'create')
        self.upload()
        self.assertEqual(listener.received[-1].payload['action'], 'update')

    def test_payload_attribute_are_sound(self):
        self.upload()
        payload = listener.received[-1].payload
        self.assertEqual(payload['uri'], self.endpoint_uri)
        self.assertEqual(payload['resource_name'], 'record')
        self.assertEqual(payload['record_id'], self.record_id)
        self.assertEqual(payload['collection_id'], 'fonts')
        self.assertEqual(payload['bucket_id'], 'fennec')



import cgi
import json
import hashlib
import gzip
from io import BytesIO

from kinto.core import utils as core_utils
from kinto.core.errors import raise_invalid
from kinto.core.storage import Filter
from kinto.views.records import Record
from kinto.authorization import RouteFactory
from pyramid import httpexceptions
from pyramid_storage.exceptions import FileNotAllowed


FILE_LINKS = '__attachments__'

RECORD_PATH = '/buckets/{bucket_id}/collections/{collection_id}/records/{id}'


class AttachmentRouteFactory(RouteFactory):
    def __init__(self, request):
        """Attachment is not a Kinto resource.

        The required permission is:
        * ``write`` if the related record exists;
        * ``record:create`` on the related collection otherwise.
        """
        super(AttachmentRouteFactory, self).__init__(request)
        self.resource_name = 'record'
        try:
            request.current_resource_name = 'record'
            request.validated.setdefault('header', {})
            request.validated.setdefault('querystring', {})
            resource = Record(request, context=self)
            resource.object_id = request.matchdict['id']
            existing = resource.get()
        except httpexceptions.HTTPNotFound:
            existing = None
        if existing:
            self.permission_object_id = record_uri(request)
            self.required_permission = 'write'
        else:
            self.permission_object_id = collection_uri(request)
            self.required_permission = 'create'


def sha256(content):
    m = hashlib.sha256()
    m.update(content)
    return m.hexdigest()


def _object_uri(request, resource_name, matchdict, prefix):
    uri = core_utils.instance_uri(request, resource_name=resource_name, **matchdict)
    if prefix:
        uri = f"/{request.registry.route_prefix}{uri}"
    return uri


def bucket_uri(request, prefix=False):
    matchdict = dict(request.matchdict)
    matchdict['id'] = matchdict['bucket_id']
    return _object_uri(request, 'bucket', matchdict, prefix)


def collection_uri(request, prefix=False):
    matchdict = dict(request.matchdict)
    matchdict['id'] = matchdict['collection_id']
    return _object_uri(request, 'collection', matchdict, prefix)


def record_uri(request, prefix=False):
    return _object_uri(request, 'record', request.matchdict, prefix)


def patch_record(record, request):
    # XXX: add util clone_request()
    backup_pattern = request.matched_route.pattern
    backup_body = request.body
    backup_validated = request.validated

    # Instantiate record resource with current request.
    context = RouteFactory(request)
    context.resource_name = 'record'
    context.get_permission_object_id = lambda r, i: record_uri(r)
    record_pattern = request.matched_route.pattern.replace('/attachment', '')
    request.matched_route.pattern = record_pattern

    # Simulate update of fields.
    request.validated = dict(body=record, **backup_validated)

    request.body = json.dumps(record).encode('utf-8')
    resource = Record(request, context=context)
    resource.object_id = request.matchdict['id']
    setattr(request, '_attachment_auto_save', True)  # Flag in update listener.

    try:
        saved = resource.patch()
    except httpexceptions.HTTPNotFound:
        saved = resource.put()

    request.matched_route.pattern = backup_pattern
    request.body = backup_body
    request.validated = backup_validated
    return saved


def delete_attachment(request, link_field=None, uri=None, keep_old_files=False):
    """Delete existing file and link."""
    if link_field is None:
        link_field = "record_uri"
    if uri is None:
        uri = record_uri(request)

    storage = request.registry.storage
    filters = [Filter(link_field, uri, core_utils.COMPARISON.EQ)]

    # Remove file.
    if not keep_old_files:
        file_links, _ = storage.get_all("", FILE_LINKS, filters=filters)
        for link in file_links:
            request.attachment.delete(link['location'])

    # Remove link.
    storage.delete_all("", FILE_LINKS, filters=filters, with_deleted=False)


def save_file(request, content, folder=None, keep_link=True, replace=False):
    gzipped = setting_value(request, 'gzipped', default=False)
    randomize = setting_value(request, 'randomize', default=True)

    # Read file to compute hash.
    if not isinstance(content, cgi.FieldStorage):
        error_msg = 'Filename is required.'
        raise_invalid(request, location='body', description=error_msg)

    # Posted file attributes.
    content.file.seek(0)
    filecontent = content.file.read()
    filehash = sha256(filecontent)
    size = len(filecontent)
    mimetype = content.type
    filename = content.filename

    original = None
    save_options = {
        'folder': folder,
        'randomize': randomize,
        'replace': replace,
        'headers': {'Content-Type': mimetype},
    }

    if gzipped:
        original = {
            'filename': filename,
            'hash': filehash,
            'mimetype': mimetype,
            'size': size,
        }
        mimetype = 'application/x-gzip'
        filename += '.gz'
        content.filename = filename
        save_options['extensions'] = ['gz']
        save_options['headers']['Content-Type'] = mimetype

        # in-memory gzipping
        out = BytesIO()
        with gzip.GzipFile(fileobj=out, mode="w") as f:
            f.write(filecontent)

        filecontent = out.getvalue()
        out.seek(0)
        content.file = out

        # We give the hash and size of the gzip content in the attachment
        # metadata.
        filehash = sha256(filecontent)
        size = len(filecontent)

    try:
        location = request.attachment.save(content, **save_options)
    except FileNotAllowed:
        error_msg = 'File extension is not allowed.'
        raise_invalid(request, location='body', description=error_msg)

    # File metadata.
    fullurl = request.attachment.url(location)
    attachment = {
        'filename': filename,
        'location': fullurl,
        'hash': filehash,
        'mimetype': mimetype,
        'size': size
    }
    if original is not None:
        attachment['original'] = original

    if keep_link:
        # Store link between record and attachment (for later deletion).
        request.registry.storage.create("", FILE_LINKS, {
            'location': location,  # store relative location.
            'bucket_uri': bucket_uri(request),
            'collection_uri': collection_uri(request),
            'record_uri': record_uri(request)
        })

    return attachment


def setting_value(request, name, default):
    value = request.registry.settings.get('attachment.{}'.format(name), default)
    if 'bucket_id' in request.matchdict:
        uri = '/buckets/{bucket_id}'.format(**request.matchdict)
        if uri in request.registry.attachment_resources:
            value = request.registry.attachment_resources[uri].get(name, value)
        if 'collection_id' in request.matchdict:
            uri = '/buckets/{bucket_id}/collections/{collection_id}'.format(**request.matchdict)
            if uri in request.registry.attachment_resources:
                value = request.registry.attachment_resources[uri].get(name, value)
    return value



import pytest
import unittest

from pyramid import testing
from pyramid.exceptions import ConfigurationError
from kinto import main as kinto_main
from kinto_attachment import __version__, includeme

from . import BaseWebTestLocal


class HelloViewTest(BaseWebTestLocal, unittest.TestCase):
    def test_capability_is_exposed(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("attachments", capabilities)
        expected = {
            "version": __version__,
            "description": "Add file attachments to records",
            "url": "https://github.com/Kinto/kinto-attachment/",
            "base_url": "https://files.server.com/root/",
        }
        self.assertEqual(expected, capabilities["attachments"])


class IncludeMeTest(unittest.TestCase):
    def includeme(self, settings):
        config = testing.setUp(settings=settings)
        kinto_main(None, config=config)
        includeme(config)
        return config

    def test_includeme_understand_authorized_resources_settings(self):
        config = self.includeme(settings={
            "attachment.base_path": "/tmp",
            "attachment.resources.fennec.gzipped": "true",
            "attachment.resources.fingerprinting.fonts.randomize": "true",
        })
        assert isinstance(config.registry.attachment_resources, dict)
        assert '/buckets/fennec' in config.registry.attachment_resources
        assert '/buckets/fingerprinting/collections/fonts' in config.registry.attachment_resources

    def test_includeme_raises_error_for_malformed_resource_settings(self):
        with pytest.raises(ConfigurationError) as excinfo:
            self.includeme(settings={"attachment.resources.fen.nec.fonts.gzipped": "true"})
        assert str(excinfo.value) == (
            'Configuration rule malformed: `attachment.resources.fen.nec.fonts.gzipped`')

    def test_includeme_raises_error_if_wrong_resource_settings_is_defined(self):
        with pytest.raises(ConfigurationError) as excinfo:
            self.includeme(settings={"attachment.resources.fennec.base_path": "foobar"})
        assert str(excinfo.value) == ('`base_path` is not a supported setting name. '
                                      'Read `attachment.resources.fennec.base_path`')



from kinto_http import cli_utils

DEFAULT_SERVER = "http://localhost:8888/v1"


def delete_attachments(client, records):
    for record in records:
        record_uri = client.get_endpoint('record', id=record['id'])
        client.session.request(method='delete',
                               endpoint=record_uri + '/attachment')

        record = client.get_record(id=record['id'])["data"]
        assert record["attachment"] is None


def main():
    parser = cli_utils.add_parser_options(
        description='Delete files from Kinto',
        default_server=DEFAULT_SERVER)

    args = parser.parse_args()

    client = cli_utils.create_client_from_args(args)
    existing = client.get_records()
    delete_attachments(client, existing)


if __name__ == '__main__':
    main()



import os
import requests
import uuid
import unittest
from unittest import mock

from urllib.parse import urlparse
from kinto.core.errors import ERRORS
from kinto_attachment.utils import sha256
from . import BaseWebTestLocal, BaseWebTestS3, get_user_headers


class UploadTest(object):
    def test_returns_200_to_record_once_uploaded(self):
        self.upload(status=201)

    def test_record_is_created_with_metadata(self):
        self.upload()
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIn(self.file_field, resp.json['data'])

    def test_record_is_created_with_valid_id(self):
        self.record_uri = self.get_record_uri('fennec', 'fonts', 'logo')
        self.endpoint_uri = self.record_uri + '/attachment'
        self.app.put_json(self.record_uri, {}, headers=self.headers)
        self.upload(status=200)

    def test_returns_200_if_record_already_exists(self):
        self.app.put_json(self.record_uri, {}, headers=self.headers)
        self.upload(status=200)

    def test_adds_cors_and_location_to_response(self):
        response = self.upload()
        self.assertEqual(response.headers['Location'],
                         'http://localhost/v1' + self.record_uri)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_has_no_subfolder_if_setting_is_undefined(self):
        self.app.app.registry.settings.pop('attachment.folder')
        response = self.upload()
        record = self.get_record(response)
        url = urlparse(record['location'])
        self.assertNotIn('/', url.path[1:])

    def exists(self, fullurl):
        location = fullurl.replace(self.base_url, '')
        return self.backend.exists(location)

    def test_previous_attachment_is_removed_on_replacement(self):
        first = self.get_record(self.upload())
        self.assertTrue(self.exists(first['location']))
        second = self.get_record(self.upload())
        self.assertFalse(self.exists(first['location']))
        self.assertTrue(self.exists(second['location']))


class LocalUploadTest(UploadTest, BaseWebTestLocal, unittest.TestCase):
    def test_file_is_created_on_local_filesystem(self):
        attachment = self.upload().json
        fullurl = attachment['location']
        relativeurl = fullurl.replace(self.base_url, '')
        self.assertTrue(os.path.exists(os.path.join('/tmp', relativeurl)))

    def test_file_is_not_gzipped_on_local_filesystem(self):
        resp = self.upload(files=[
            (self.file_field, b'my-report.pdf', b'--binary--')
        ])
        attachment = resp.json
        self.assertTrue(attachment['location'].endswith('.pdf'))
        self.assertEqual(attachment['mimetype'], 'application/pdf')
        relativeurl = attachment['location'].replace(self.base_url, '')
        self.assertEqual(attachment['hash'], sha256(b'--binary--'))
        self.assertEqual(attachment['size'], len(b'--binary--'))
        file_path = os.path.join('/tmp', relativeurl)
        self.assertTrue(os.path.exists(file_path))
        with open(file_path, 'rb') as f:
            self.assertEqual(f.read(), b'--binary--')


class S3UploadTest(UploadTest, BaseWebTestS3, unittest.TestCase):
    pass


class DeleteTest(object):
    def setUp(self):
        super(DeleteTest, self).setUp()
        self.attachment = self.upload().json

    def exists(self, fullurl):
        location = fullurl.replace(self.base_url, '')
        return self.backend.exists(location)

    def test_attachment_is_removed_on_delete(self):
        fullurl = self.attachment['location']
        self.assertTrue(self.exists(fullurl))
        self.app.delete(self.endpoint_uri, headers=self.headers, status=204)
        self.assertFalse(self.exists(fullurl))

    def test_metadata_are_removed_on_delete(self):
        self.app.delete(self.endpoint_uri, headers=self.headers, status=204)
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIsNone(resp.json['data'].get('attachment'))

    def test_link_is_removed_on_delete(self):
        storage = self.app.app.registry.storage
        links, _ = storage.get_all("", '__attachments__')
        self.assertEqual(len(links), self.nb_uploaded_files)
        self.app.delete(self.endpoint_uri, headers=self.headers, status=204)
        links, _ = storage.get_all("", '__attachments__')
        self.assertEqual(len(links), 0)

    def test_attachment_is_removed_when_record_is_deleted(self):
        fullurl = self.attachment['location']
        self.assertTrue(self.exists(fullurl))
        self.app.delete(self.record_uri, headers=self.headers)
        self.assertFalse(self.exists(fullurl))

    def test_attachments_are_removed_when_bucket_is_deleted(self):
        fullurl = self.attachment['location']
        self.assertTrue(self.exists(fullurl))
        self.app.delete('/buckets/fennec', headers=self.headers)
        self.assertFalse(self.exists(fullurl))

    def test_attachments_are_removed_when_collection_is_deleted(self):
        fullurl = self.attachment['location']
        self.assertTrue(self.exists(fullurl))
        self.app.delete('/buckets/fennec/collections/fonts',
                        headers=self.headers)
        self.assertFalse(self.exists(fullurl))

    def test_attachments_links_are_removed_forever(self):
        storage = self.app.app.registry.storage
        links, _ = storage.get_all("", '__attachments__')
        self.assertEqual(len(links), self.nb_uploaded_files)
        self.app.delete(self.record_uri, headers=self.headers)
        links, _ = storage.get_all("", '__attachments__')
        self.assertEqual(len(links), 0)

    def test_no_error_when_other_resource_is_deleted(self):
        group_url = '/buckets/default/groups/admins'
        self.app.put_json(group_url, {"data": {"members": ["them"]}},
                          headers=self.headers)
        self.app.delete(group_url, headers=self.headers)


class LocalDeleteTest(DeleteTest, BaseWebTestLocal, unittest.TestCase):
    pass


class S3DeleteTest(DeleteTest, BaseWebTestS3, unittest.TestCase):
    pass


class AttachmentViewTest(object):

    def test_only_post_and_options_is_accepted(self):
        self.app.get(self.endpoint_uri, headers=self.headers, status=405)
        self.app.put(self.endpoint_uri, headers=self.headers, status=405)
        self.app.patch(self.endpoint_uri, headers=self.headers, status=405)
        headers = self.headers.copy()
        headers['Access-Control-Request-Method'] = 'POST'
        self.app.options(self.endpoint_uri, headers=headers, status=200)

    def test_record_is_updated_with_metadata(self):
        existing = {'data': {'author': 'frutiger'}}
        self.app.put_json(self.record_uri, existing, headers=self.headers)
        self.upload()
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIn(self.file_field, resp.json['data'])
        self.assertIn('author', resp.json['data'])

    def test_record_metadata_has_hash_hexdigest(self):
        r = self.upload()
        h = 'db511d372e98725a61278e90259c7d4c5484fc7a781d7dcc0c93d53b8929e2ba'
        self.assertEqual(self.get_record(r)['hash'], h)

    def test_record_metadata_has_randomized_location(self):
        resp = self.upload(files=[
            (self.file_field, b'my-report.pdf', b'--binary--')
        ])
        record = self.get_record(resp)
        self.assertNotIn('report', record['location'])

    def test_record_location_contains_subfolder(self):
        self.upload()
        resp = self.app.get(self.record_uri, headers=self.headers)
        location = resp.json['data'][self.file_field]['location']
        self.assertIn('fennec/fonts/', location)

    def test_record_metadata_provides_original_filename(self):
        resp = self.upload(files=[
            (self.file_field, b'my-report.pdf', b'--binary--')
        ])
        record = self.get_record(resp)
        self.assertEqual('my-report.pdf', record['filename'])

    def test_record_is_created_with_fields(self):
        self.upload(params=[('data', '{"family": "sans"}')])
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertEqual(resp.json['data']['family'], "sans")

    def test_record_is_updated_with_fields(self):
        existing = {'data': {'author': 'frutiger'}}
        self.app.put_json(self.record_uri, existing, headers=self.headers)
        self.upload(params=[('data', '{"family": "sans"}')])
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertEqual(resp.json['data']['family'], 'sans')
        self.assertEqual(resp.json['data']['author'], 'frutiger')

    def test_record_attachment_metadata_cannot_be_removed_manually(self):
        self.upload(params=[('data', '{"family": "sans"}')])
        body = {'data': {'attachment': {'manual': 'true'}}}
        resp = self.app.patch_json(self.record_uri, body, headers=self.headers,
                                   status=400)
        self.assertIn('Attachment metadata cannot be modified',
                      resp.json['message'])

    def test_record_is_created_with_appropriate_permissions(self):
        self.upload()
        current_principal = ("basicauth:c6c27f0c7297ba7d4abd2a70c8a2cb88a06a3"
                             "bb793817ef2c85fe8a709b08022")
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertEqual(resp.json['permissions'],
                         {"write": [current_principal]})

    def test_record_permissions_can_also_be_specified(self):
        self.upload(params=[('permissions', '{"read": ["system.Everyone"]}')])
        resp = self.app.get(self.record_uri, headers=self.headers)
        self.assertIn('system.Everyone', resp.json['permissions']['read'])

    # Content Validation.

    def test_records_fields_must_be_valid_json(self):
        resp = self.upload(params=[('data', '{>author: 12}')], status=400)
        self.assertIn('body: data is not valid JSON', resp.json['message'])

    def test_permissions_must_be_valid_json(self):
        resp = self.upload(params=[('permissions', '{"read": >}')], status=400)
        self.assertIn('body: permissions is not valid JSON',
                      resp.json['message'])

    def test_unknown_fields_are_not_accepted(self):
        resp = self.upload(params=[('my_field', 'a_value')], status=400)
        self.assertIn("body: 'my_field' not in ('data', 'permissions')",
                      resp.json['message'])

    def test_record_fields_are_validated_against_schema(self):
        resp = self.upload(params=[('data', '{"author": 12}')], status=400)
        self.assertIn("author in body: 12 is not of type ", resp.json['message'])

    def test_attachment_must_have_a_filename(self):
        resp = self.upload(files=[(self.file_field, b'', b'--fake--')],
                           status=400)
        self.assertEqual(resp.json['message'],
                         'body: Filename is required.')

    def test_upload_refused_if_extension_not_allowed(self):
        resp = self.upload(files=[(self.file_field, b'virus.exe',
                                   b'--fake--')], status=400)
        self.assertEqual(resp.json['message'],
                         'body: File extension is not allowed.')

    def test_upload_refused_if_field_is_not_attachment(self):
        resp = self.upload(files=[('fichierjoint', b'image.jpg', b'--fake--')],
                           status=400)
        self.assertEqual(resp.json['message'],
                         'Attachment missing.')
        self.assertEqual(resp.json['errno'], ERRORS.INVALID_POSTED_DATA.value)

    def test_upload_refused_if_header_is_not_multipart(self):
        self.headers['Content-Type'] = 'application/json'
        resp = self.app.post(self.endpoint_uri, {},
                             headers=self.headers,
                             status=400)
        self.assertEqual(resp.json['message'],
                         "Content-Type should be multipart/form-data")
        self.assertEqual(resp.json['errno'], ERRORS.INVALID_PARAMETERS.value)

    def test_upload_refused_if_header_is_invalid_multipart(self):
        self.headers['Content-Type'] = 'multipart/form-data'
        resp = self.app.post(self.endpoint_uri, {},
                             headers=self.headers,
                             status=400)
        self.assertEqual(resp.json['message'].replace(": b'", ": '"),
                         "Invalid boundary in multipart form: ''")
        self.assertEqual(resp.json['errno'], ERRORS.INVALID_PARAMETERS.value)

    # Permissions.

    def test_upload_refused_if_not_authenticated(self):
        self.headers.pop('Authorization')
        self.upload(status=401)

    def test_upload_refused_if_not_allowed(self):
        self.headers.update(get_user_headers('jean-louis'))
        self.upload(status=403)

    def test_upload_replace_refused_if_only_create_allowed(self):
        # Allow any authenticated to write in this bucket.
        perm = {'permissions': {'record:create': ['system.Authenticated']}}
        self.app.patch_json('/buckets/fennec/collections/fonts',
                            perm, headers=self.headers)
        self.upload(status=201)

        self.headers.update(get_user_headers('jean-louis'))
        self.upload(status=403)

    def test_upload_create_accepted_if_create_allowed(self):
        # Allow any authenticated to write in this bucket.
        perm = {'permissions': {'record:create': ['system.Authenticated']}}
        self.app.patch_json('/buckets/fennec/collections/fonts',
                            perm, headers=self.headers)

        self.headers.update(get_user_headers('jean-louis'))
        self.upload(status=201)

    def test_upload_create_accepted_if_write_allowed(self):
        # Allow any authenticated to write in this bucket.
        perm = {'permissions': {'write': ['system.Authenticated']}}
        self.app.patch_json('/buckets/fennec', perm, headers=self.headers)

        self.headers.update(get_user_headers('jean-louis'))
        self.upload(status=201)

    def test_upload_replace_accepted_if_write_allowed(self):
        # Allow any authenticated to write in this bucket.
        perm = {'permissions': {'write': ['system.Authenticated']}}
        self.app.patch_json('/buckets/fennec', perm, headers=self.headers)
        self.upload(status=201)

        self.headers.update(get_user_headers('jean-louis'))
        self.upload(status=200)


class ZippedAttachementViewTest(BaseWebTestLocal, unittest.TestCase):
    config = 'config/local_gzipped.ini'

    def test_file_get_zipped_if_configured(self):
        r = self.upload()
        self.assertEqual(r.json['mimetype'], 'application/x-gzip')
        self.assertEqual(r.json['filename'], 'image.jpg.gz')


class PerResourceConfigAttachementViewTest(BaseWebTestS3, unittest.TestCase):
    config = 'config/s3_per_resource.ini'

    def test_file_get_zipped_in_fennec_bucket(self):
        r = self.upload()

        self.assertEqual(r.json['original']['mimetype'], 'image/jpeg')
        self.assertEqual(r.json['mimetype'], 'application/x-gzip')
        self.assertEqual(r.json['filename'], 'image.jpg.gz')

        relative_url = r.json['location'].replace(self.base_url, '')
        resp = requests.get("http://localhost:5000/myfiles/{}".format(relative_url))
        self.assertEqual(resp.headers['Content-Type'], 'application/octet-stream')
        self.assertNotIn('Content-Encoding', resp.headers)

    def test_file_do_not_get_zipped_in_fennec_experiments_collection(self):
        self.create_collection('fennec', 'experiments')
        record_uri = self.get_record_uri('fennec', 'experiments', str(uuid.uuid4()))
        self.endpoint_uri = record_uri + '/attachment'
        r = self.upload()

        self.assertNotIn('original', r.json['mimetype'])
        self.assertEqual(r.json['mimetype'], 'image/jpeg')


class SingleAttachmentViewTest(AttachmentViewTest, BaseWebTestLocal,
                               unittest.TestCase):
    pass


class DefaultBucketTest(BaseWebTestLocal, unittest.TestCase):
    def setUp(self):
        super(DefaultBucketTest, self).setUp()
        self.record_uri = self.get_record_uri('default', 'pix', uuid.uuid4())
        self.endpoint_uri = self.record_uri + '/attachment'

    def test_implicit_collection_creation_on_upload(self):
        resp = self.upload()
        record_uri = resp.headers['Location']
        self.assertIn('/buckets/c0343679-10aa-a101-bf0f-e96f917f3e27',
                      record_uri)


class KeepOldFilesTest(BaseWebTestLocal, unittest.TestCase):
    def make_app(self):
        import webtest
        from kinto import main as testapp
        from kinto import DEFAULT_SETTINGS
        from kinto.core import testing as core_support

        settings = core_support.DEFAULT_SETTINGS.copy()
        settings.update(**DEFAULT_SETTINGS)
        settings['multiauth.policies'] = 'basicauth'
        settings['storage_backend'] = 'kinto.core.storage.memory'
        settings['permission_backend'] = 'kinto.core.permission.memory'
        settings['userid_hmac_secret'] = "this is not a secret"
        settings['includes'] = "kinto_attachment"

        settings['kinto.attachment.base_path'] = "/tmp"
        settings['kinto.attachment.base_url'] = ""
        settings['kinto.attachment.keep_old_files'] = "true"

        app = webtest.TestApp(testapp({}, **settings))
        app.RequestClass = core_support.get_request_class(prefix="v1")
        return app

    def test_files_are_kept_when_attachment_is_replaced(self):
        resp = self.upload(status=201)
        location1 = resp.json["location"]
        resp = self.upload(status=200)
        location2 = resp.json["location"]
        self.assertNotEqual(location1, location2)
        self.assertTrue(self.backend.exists(location2))
        self.assertTrue(self.backend.exists(location1))

    def test_files_are_kept_when_attachment_is_deleted(self):
        resp = self.upload(status=201)
        location = resp.json["location"]
        self.assertTrue(self.backend.exists(location))

        self.app.delete(self.record_uri + "/attachment", headers=self.headers)

        self.assertTrue(self.backend.exists(location))

    def test_files_are_kept_when_record_is_deleted(self):
        resp = self.upload(status=201)
        location = resp.json["location"]
        self.assertTrue(self.backend.exists(location))

        self.app.delete(self.record_uri, headers=self.headers)

        self.assertTrue(self.backend.exists(location))


class HeartbeartTest(BaseWebTestS3, unittest.TestCase):
    def test_attachments_is_added_to_heartbeat_view(self):
        resp = self.app.get('/__heartbeat__')
        self.assertIn('attachments', resp.json)

    def test_heartbeat_is_false_if_error_happens(self):
        with mock.patch('pyramid_storage.s3.S3FileStorage.delete') as mocked:
            mocked.side_effect = ValueError
            resp = self.app.get('/__heartbeat__', status=503)
        self.assertFalse(resp.json['attachments'])

    def test_heartbeat_is_true_if_server_is_readonly(self):
        patch = mock.patch('pyramid_storage.s3.S3FileStorage.delete')
        self.addCleanup(patch.stop)
        mocked = patch.start()
        mocked.side_effect = ValueError

        with mock.patch.dict(self.app.app.registry.settings,
                             [('readonly', 'true')]):
            resp = self.app.get('/__heartbeat__')
        self.assertTrue(resp.json['attachments'])



import json
import hashlib
import mimetypes
import os
import pprint
import uuid

from kinto_http import cli_utils
from kinto_http.exceptions import KintoException

DEFAULT_SERVER = "http://localhost:8888/v1"


def sha256(content):
    m = hashlib.sha256()
    m.update(content)
    return m.hexdigest()


def files_to_upload(records, files, force=False):
    records_by_id = {r['id']: r for r in records if 'attachment' in r}
    existing_files = {r['attachment']['filename']: r for r in records if 'attachment' in r}
    existing_original_files = {r['attachment']['original']['filename']: r
                               for r in records
                               if 'attachment' in r and 'original' in r['attachment']}
    to_upload = []
    for filepath in files:
        filename = os.path.basename(filepath)

        record = None
        if filename in existing_files.keys():
            record = existing_files[filename]
        elif filename in existing_original_files.keys():
            record = existing_original_files[filename]

        if record:
            records_by_id.pop(record['id'], None)
            local_hash = sha256(open(filepath, 'rb').read())

            # If file was uploaded gzipped, compare with hash of
            # uncompressed file.
            remote_hash = record.get('original', {}).get('hash')
            if not remote_hash:
                remote_hash = record['attachment']['hash']

            # If hash has changed, upload !
            if local_hash != remote_hash or force:
                print("File '%s' has changed." % filename)
                to_upload.append((filepath, record))
            else:
                print("File '%s' is up-to-date." % filename)
        else:
            identifier = hashlib.md5(filename.encode('utf-8')).hexdigest()
            record_id = str(uuid.UUID(identifier))
            record = {'id': record_id}
            to_upload.append((filepath, record))

    # XXX: add option to delete records when files are missing locally
    for id, record in records_by_id.items():
        print("Ignore remote file '%s'." % record['attachment']['filename'])

    return to_upload


def upload_files(client, files):
    permissions = {}  # XXX: Permissions are inherited from collection.

    for filepath, record in files:
        mimetype, _ = mimetypes.guess_type(filepath)
        filename = os.path.basename(filepath)
        filecontent = open(filepath, "rb").read()
        record_uri = client.get_endpoint('record', id=record['id'])
        attachment_uri = '%s/attachment' % record_uri
        multipart = [("attachment", (filename, filecontent, mimetype))]
        try:
            body, _ = client.session.request(method='post',
                                             endpoint=attachment_uri,
                                             permissions=json.dumps(permissions),
                                             files=multipart)
        except KintoException as e:
            print(filepath, "error during upload.", e)
        else:
            pprint.pprint({"id": record['id'], "attachment": body})


def main():
    parser = cli_utils.add_parser_options(
        description='Upload files to Kinto',
        default_server=DEFAULT_SERVER)

    parser.add_argument('--force', dest='force', action='store_true',
                        help='Force upload even if the hash matches')
    parser.add_argument('files', metavar='FILE', action='store',
                        nargs='+')
    args = parser.parse_args()

    client = cli_utils.create_client_from_args(args)

    try:
        client.create_bucket(if_not_exists=True)
        client.create_collection(if_not_exists=True)
    except KintoException:
        # Fail silently in case of 403
        pass

    existing = client.get_records()
    to_upload = files_to_upload(existing, args.files, force=args.force)
    upload_files(client, to_upload)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    readme = f.read()

with codecs.open(os.path.join(here, 'CHANGELOG.rst'), encoding='utf-8') as f:
    history = f.read()

requirements = [
    'boto',
    'kinto>=5.1.0',
    'pyramid_storage>=0.1.0',
]

test_requirements = [
    'webtest'
]

setup(
    name='kinto-attachment',
    version='6.0.4',
    description="Attach files to Kinto records",
    long_description=readme + '\n\n' + history,
    author="Mozilla",
    author_email='kinto@mozilla.org',
    url='https://github.com/Kinto/kinto-attachment',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="Apache License (2.0)",
    zip_safe=False,
    keywords='kinto',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='kinto_attachment.tests',
    tests_require=test_requirements
)



import sys
from kinto_client import cli_utils


DEFAULT_SERVER = "https://kinto-ota.dev.mozaws.net/v1/"


def _print(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()


def upgrade(client):
    _print('Scanning collection...\n')
    count = 0

    for record in client.get_records():
        _print('.')

        attachment = record.get('attachment')
        if attachment is None:
            continue

        original = attachment.get('original')
        old_original = record.get('original')

        if original is None and old_original is not None:
            record['attachment']['original'] = old_original
            del record['original']
            client.update_record(record)
            count += 1

    _print('\nChanged %d records\n' % count)


def main():
    parser = cli_utils.add_parser_options(
        description='Upgrade the attachment structure',
        default_server=DEFAULT_SERVER)
    args = parser.parse_args()
    client = cli_utils.create_client_from_args(args)
    upgrade(client)


if __name__ == '__main__':
    main()
